"""
The MIT License

Copyright 2019 Derek Miller

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions
of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import socket
from multiprocessing import Process
from flask import Flask, request
import dill
from sparktorch.util import load_torch_model, TorchObj
from sparktorch.rw_lock import RWLock

import logging
log = logging.getLogger('werkzeug')
log.setLevel(logging.ERROR)


class Server(object):

    def __init__(
        self,
        torch_obj: TorchObj,
        master_url: str = None,
        port: int = 3000,
        acquire_lock: bool = True
    ):
        self.torch_obj = load_torch_model(torch_obj)

        self.model = self.torch_obj.model

        self.state_dict = self.model.state_dict()
        self.criterion = self.torch_obj.criterion
        self.optimizer = self.torch_obj.optimizer

        self.master_url = master_url
        self.port = port
        self.error_count = 0
        self.acquire_lock = acquire_lock

        self.server = Process(target=self.start_service)

    @staticmethod
    def determine_master(port: int):
        try:
            master_url = socket.gethostbyname(socket.gethostname()) + ':' + str(port)
            return master_url
        except:
            return 'localhost:' + str(port)

    def start_server(self):
        self.server.start()
        self.master_url = Server.determine_master(self.port)

    def stop_server(self):
        self.server.terminate()
        self.server.join()

    def start_service(self):
        app = Flask(__name__)
        self.app = app
        self.model.train()
        lock = RWLock()
        lock_acquired = self.acquire_lock

        @app.route('/')
        def home():
            return 'sparktorch'

        @app.route('/parameters', methods=['GET'])
        def get_parameters():
            if lock_acquired:
                lock.acquire_read()
            state = dill.dumps(self.state_dict)
            if lock_acquired:
                lock.release()
            return state

        @app.route('/update', methods=['POST'])
        def update_parameters():

            if lock_acquired:
                lock.acquire_write()

            try:
                gradients = dill.loads(request.data)

                for index, param in enumerate(self.model.parameters()):
                    param.grad = gradients[index]

                self.optimizer.step()

                self.state_dict = self.model.state_dict()

            except Exception as e:
                self.error_count += 1
                if self.error_count > 10:
                    raise RuntimeError(f"Max Errors {str(e)}")
            finally:
                if lock_acquired:
                    lock.release()

            return 'completed'

        self.app.run(host='0.0.0.0', use_reloader=False, threaded=True, port=self.port)





"""
The MIT License

Copyright 2019 Derek Miller

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions
of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import numpy as np
import torch
import torch.nn as nn
from torch.optim.optimizer import Optimizer
import dill
from typing import Any, Dict, List, Type
import collections
import codecs

TorchObj = collections.namedtuple('TorchObj', ['model', 'criterion', 'optimizer', 'optimizer_params'])

DataObj = collections.namedtuple('DataObj', ['x_train', 'y_train', 'x_val', 'y_val'])


def torch_encoder(obj):
    """
    Encodes Torch Object or anything related
    :param obj: any object
    :return: decoded object
    """
    return codecs.encode(
        dill.dumps(obj), "base64"
    ).decode()


def torch_decoder(model_ser):
    return dill.loads(codecs.decode(model_ser.encode(), "base64"))


def handle_features(data: List[DataObj]) -> DataObj:
    x_train = []
    y_train = []

    for feature in data:

        if feature.y_train is not None:
            if type(feature.y_train) is int or type(feature.y_train) is float:
                y_train.append([feature.y_train])
            else:
                y_train.append(feature.y_train)

        x_train.append(feature.x_train)

    x_train = torch.from_numpy(np.stack(x_train)).float()
    y_train = torch.from_numpy(np.asarray(y_train)).float() if y_train else None

    return DataObj(x_train=x_train, y_train=y_train, x_val=None, y_val=None)


def load_torch_model(torch_obj: str) -> TorchObj:
    loaded = torch_decoder(torch_obj)
    model = loaded.model
    optimizer = load_optimizer(loaded.optimizer, loaded.optimizer_params, model)
    return TorchObj(
        model=model,
        criterion=loaded.criterion,
        optimizer=optimizer,
        optimizer_params=loaded.optimizer_params
    )


def serialize_torch_obj(
        model: nn.Module,
        criterion: Any,
        optimizer: Type[Optimizer],
        **kwargs
) -> str:
    return torch_encoder(
        TorchObj(
            model=model,
            criterion=criterion,
            optimizer=optimizer,
            optimizer_params=kwargs
        )
    )


def load_optimizer(optimizer: Type[Optimizer], params: Dict, model: nn.Module) -> Optimizer:
    if params:
        return optimizer(model.parameters(), **params)

    return optimizer(model.parameters())



import dill
from pyspark.ml import Pipeline, PipelineModel
from pyspark.ml.util import JavaMLReader, JavaMLWriter
from pyspark.ml.feature import StopWordsRemover
from pyspark.ml.wrapper import JavaParams
from pyspark.context import SparkContext
import zlib
import sys

"""
based off below stackoverflow thread. Changes were made for performance.
credit: https://stackoverflow.com/questions/41399399/serialize-a-custom-transformer-using-python-to-be-used-within-a-pyspark-ml-pipel
"""


class PysparkObjId(object):
    """
    A class to specify constants used to idenify and setup python
    Estimators, Transformers and Models so they can be serialized on there
    own and from within a Pipline or PipelineModel.
    """
    def __init__(self):
        super(PysparkObjId, self).__init__()

    @staticmethod
    def _getPyObjId():
        return '4c1740b00d3c4ff6806a1402321572cb'

    @staticmethod
    def _getCarrierClass(javaName=False):
        return 'org.apache.spark.ml.feature.StopWordsRemover' if javaName else StopWordsRemover


def load_byte_array(stop_words):
    swords = stop_words[0].split(',')[0:-1]
    if sys.version_info[0] < 3:
        lst = [chr(int(d)) for d in swords]
        dmp = ''.join(lst)
        dmp = zlib.decompress(dmp)
        py_obj = dill.loads(dmp)
        return py_obj
    dmp = bytes([int(i) for i in swords])
    dmp = zlib.decompress(dmp)
    py_obj = dill.loads(dmp)
    return py_obj


class PysparkPipelineWrapper(object):
    """
    A class to facilitate converting the stages of a Pipeline or PipelineModel
    that were saved from PysparkReaderWriter.
    """
    def __init__(self):
        super(PysparkPipelineWrapper, self).__init__()

    @staticmethod
    def unwrap(pipeline):
        if not (isinstance(pipeline, Pipeline) or isinstance(pipeline, PipelineModel)):
            raise TypeError("Cannot recognize a pipeline of type %s." % type(pipeline))

        stages = pipeline.getStages() if isinstance(pipeline, Pipeline) else pipeline.stages
        for i, stage in enumerate(stages):
            if (isinstance(stage, Pipeline) or isinstance(stage, PipelineModel)):
                stages[i] = PysparkPipelineWrapper.unwrap(stage)
            if isinstance(stage, PysparkObjId._getCarrierClass()) and stage.getStopWords()[-1] == PysparkObjId._getPyObjId():
                swords = stage.getStopWords()[:-1] # strip the id
                py_obj = load_byte_array(swords)
                stages[i] = py_obj

        if isinstance(pipeline, Pipeline):
            pipeline.setStages(stages)
        else:
            pipeline.stages = stages
        return pipeline


class PysparkReaderWriter(object):
    """
    A mixin class so custom pyspark Estimators, Transformers and Models may
    support saving and loading directly or be saved within a Pipline or PipelineModel.
    """
    def __init__(self):
        super(PysparkReaderWriter, self).__init__()

    def write(self):
        """Returns an MLWriter instance for this ML instance."""
        return JavaMLWriter(self)

    @classmethod
    def read(cls):
        """Returns an MLReader instance for our clarrier class."""
        return JavaMLReader(PysparkObjId._getCarrierClass())

    @classmethod
    def load(cls, path):
        """Reads an ML instance from the input path, a shortcut of `read().load(path)`."""
        swr_java_obj = cls.read().load(path)
        return cls._from_java(swr_java_obj)

    @classmethod
    def _from_java(cls, java_obj):
        """
        Get the dumby the stopwords that are the characters of the dills dump plus our guid
        and convert, via dill, back to our python instance.
        """
        swords = java_obj.getStopWords()[:-1] # strip the id
        return load_byte_array(swords)

    def _to_java(self):
        """
        Convert this instance to a dill dump, then to a list of strings with the unicode integer values of each character.
        Use this list as a set of dumby stopwords and store in a StopWordsRemover instance
        :return: Java object equivalent to this instance.
        """
        dmp = dill.dumps(self)
        dmp = zlib.compress(dmp)
        sc = SparkContext._active_spark_context
        pylist = [str(i) + ',' for i in bytearray(dmp)]
        # convert bytes to string integer list
        pylist = [''.join(pylist)]
        pylist.append(PysparkObjId._getPyObjId()) # add our id so PysparkPipelineWrapper can id us.
        java_class = sc._gateway.jvm.java.lang.String
        java_array = sc._gateway.new_array(java_class, len(pylist))
        java_array[0:2] = pylist[0:2]
        _java_obj = JavaParams._new_java_obj(PysparkObjId._getCarrierClass(javaName=True), self.uid)
        _java_obj.setStopWords(java_array)
        return _java_obj




"""
The MIT License

Copyright 2019 Derek Miller

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions
of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

from sparktorch.pipeline_util import PysparkReaderWriter
from sparktorch.util import DataObj, load_torch_model
from sparktorch.server import Server
from sparktorch.hogwild import train, get_main
import numpy as np
import torch
import codecs

from pyspark.ml.param import Param, Params, TypeConverters
from pyspark.ml.param.shared import HasInputCol, HasPredictionCol, HasLabelCol
from pyspark.ml.base import Estimator
from pyspark.ml import Model
from pyspark.ml.util import Identifiable, MLReadable, MLWritable
from pyspark import keyword_only
import time
from pyspark.ml.linalg import Vectors, VectorUDT
from pyspark.sql import functions as F
import dill

from pyspark import SparkContext


def handle_data(data, inp_col, label_col):
    return DataObj(
        np.asarray(data[inp_col]),
        data[label_col],
        None,
        None
    )


class SparkTorchModel(Model, HasInputCol, HasPredictionCol, PysparkReaderWriter, MLReadable, MLWritable, Identifiable):

    modStr = Param(Params._dummy(), "modStr", "", typeConverter=TypeConverters.toString)

    @keyword_only
    def __init__(
        self,
        inputCol=None,
        predictionCol=None,
        modStr=None
    ):
        super().__init__()
        self._setDefault(
            inputCol='encoded',
            predictionCol='predicted',
            modStr=''
        )
        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    @keyword_only
    def setParams(
        self,
        inputCol=None,
        predictionCol=None,
        modStr=None
    ):
        kwargs = self._input_kwargs
        return self._set(**kwargs)

    def _transform(self, dataset):
        inp = self.getOrDefault(self.inputCol)
        out = self.getOrDefault(self.predictionCol)
        mod_str = self.getOrDefault(self.modStr)
        model = dill.loads(codecs.decode(mod_str.encode(), "base64"))
        model_broadcast = dataset._sc.broadcast(model)

        def predict_func(data):
            features = np.asarray(data).reshape((1, len(data)))
            x_data = torch.from_numpy(features).float()
            model = model_broadcast.value
            model.eval()
            return Vectors.dense(model(x_data).detach().numpy().flatten())

        udfGenerateCode = F.udf(predict_func, VectorUDT())

        return dataset.withColumn(out, udfGenerateCode(inp))


class SparkTorch(
    Estimator,
    HasInputCol,
    HasPredictionCol,
    HasLabelCol,
    PysparkReaderWriter,
    MLReadable,
    MLWritable,
    Identifiable
):

    torchObj = Param(Params._dummy(), "torchObj", "", typeConverter=TypeConverters.toString)
    iters = Param(Params._dummy(), "iters", "", typeConverter=TypeConverters.toInt)
    partitions = Param(Params._dummy(), "partitions", "", typeConverter=TypeConverters.toInt)
    verbose = Param(Params._dummy(), "verbose", "", typeConverter=TypeConverters.toInt)
    acquireLock = Param(Params._dummy(), "acquireLock", "", typeConverter=TypeConverters.toBoolean)
    shufflePerIter = Param(Params._dummy(), "shufflePerIter", "", typeConverter=TypeConverters.toBoolean)
    partitionShuffles = Param(Params._dummy(), "partitionShuffles", "", typeConverter=TypeConverters.toInt)
    port = Param(Params._dummy(), "port", "", typeConverter=TypeConverters.toInt)

    @keyword_only
    def __init__(
        self,
        inputCol=None,
        labelCol=None,
        torchObj=None,
        iters=None,
        predictionCol=None,
        partitions=None,
        acquireLock=None,
        shufflePerIter=None,
        verbose=None,
        partitionShuffles=None,
        port=None
    ):
        super().__init__()
        self._setDefault(
            inputCol='features',
            labelCol=None,
            torchObj='',
            iters=10,
            predictionCol='predicted',
            partitions=-1,
            acquireLock=True,
            shufflePerIter=True,
            verbose=0,
            partitionShuffles=1,
            port=3000
        )
        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    @keyword_only
    def setParams(
        self,
        inputCol=None,
        labelCol=None,
        torchObj=None,
        iters=None,
        predictionCol=None,
        partitions=None,
        acquireLock=None,
        shufflePerIter=None,
        verbose=None,
        partitionShuffles=None,
        port=None
    ):
        kwargs = self._input_kwargs
        return self._set(**kwargs)

    def getTorchObj(self):
        return self.getOrDefault(self.torchObj)

    def getIters(self):
        return self.getOrDefault(self.iters)

    def getPartitions(self):
        return self.getOrDefault(self.partitions)

    def getVerbose(self):
        return self.getOrDefault(self.verbose)

    def getAqcuireLock(self):
        return self.getOrDefault(self.acquireLock)

    def getShufflePerIter(self):
        return self.getOrDefault(self.shufflePerIter)

    def getPartitionShuffles(self):
        return self.getOrDefault(self.partitionShuffles)

    def getPort(self):
        return self.getOrDefault(self.port)

    def _fit(self, dataset):
        inp_col = self.getInputCol()
        label = self.getLabelCol()
        prediction = self.getPredictionCol()

        torch_obj = self.getTorchObj()
        iters = self.getIters()
        partitions = self.getPartitions()
        acquire_lock = self.getAqcuireLock()
        verbose = self.getVerbose()
        spi = self.getShufflePerIter()
        partition_shuffles = self.getPartitionShuffles()
        port = self.getPort()

        df = dataset.rdd.map(lambda x: handle_data(x, inp_col, label))
        if partitions > 0:
            df = df.coalesce(partitions) if partitions < df.getNumPartitions() else df

        master_url = SparkContext._active_spark_context.getConf().get("spark.driver.host").__str__() + ":" + str(port)
        server = Server(
            torch_obj=torch_obj,
            master_url=master_url,
            port=port,
            acquire_lock=acquire_lock
        )

        server.start_server()
        time.sleep(5)
        print(f'Server is running {get_main(master_url)}')

        state_dict = train(df, torch_obj, server, iters, partition_shuffles, verbose=verbose)
        loaded = load_torch_model(torch_obj)
        model = loaded.model
        model.load_state_dict(state_dict)
        dumped_model = codecs.encode(dill.dumps(model), "base64").decode()

        return SparkTorchModel(
            inputCol=inp_col,
            predictionCol=prediction,
            modStr=dumped_model
        )



from pyspark.sql import SparkSession
import unittest
import logging
import numpy as np
from pyspark.ml.linalg import Vectors
import torch.nn as nn
import torch
from sparktorch.util import serialize_torch_obj
from sparktorch.torch_async import SparkTorch
from tests.simple_net import Net


class PysparkTest(unittest.TestCase):

    @classmethod
    def suppress_logging(cls):
        logger = logging.getLogger('py4j')
        logger.setLevel(logging.WARN)

    @classmethod
    def create_testing_spark_session(cls):
        return (SparkSession.builder
                .master('local[2]')
                .appName('sparktorch')
                .getOrCreate())

    @classmethod
    def setUpClass(cls):
        cls.suppress_logging()
        cls.spark = cls.create_testing_spark_session()

    @classmethod
    def tearDownClass(cls):
        cls.spark.stop()


class SparkTorchTests(PysparkTest):

    def generate_random_data(self):
        dat = [(1.0, Vectors.dense(np.random.normal(0,1,10))) for _ in range(0, 200)]
        dat2 = [(0.0, Vectors.dense(np.random.normal(2,1,10))) for _ in range(0, 200)]
        dat.extend(dat2)
        return self.spark.createDataFrame(dat, ["label", "features"])

    @staticmethod
    def generate_sequential_model():
        model = torch.nn.Sequential(
            nn.Linear(10, 20),
            nn.ReLU(),
            nn.Linear(20, 1)
        )
        return serialize_torch_obj(
            model, nn.MSELoss(), torch.optim.Adam, lr=0.001
        )

    def test_simple_sequential(self):
        model = self.generate_sequential_model()
        data = self.generate_random_data()

        stm = SparkTorch(
            inputCol='features',
            labelCol='label',
            predictionCol='predictions',
            torchObj=model,
            iters=5
        ).fit(data)

        res = stm.transform(data).take(1)
        self.assertTrue('predictions' in res[0])
        self.assertTrue(len(res[0]['predictions']) == 1)

    def test_simple_torch_module(self):
        model = serialize_torch_obj(
            Net(), nn.MSELoss(), torch.optim.Adam, lr=0.001
        )
        data = self.generate_random_data()

        stm = SparkTorch(
            inputCol='features',
            labelCol='label',
            predictionCol='predictions',
            torchObj=model,
            iters=5,
            verbose=1
        ).fit(data)

        res = stm.transform(data).take(1)
        self.assertTrue('predictions' in res[0])
        self.assertTrue(len(res[0]['predictions']) == 1)


if __name__ == '__main__':
    unittest.main()



from setuptools import setup
from setuptools import find_packages

setup(name='sparktorch',
      version='0.0.2',
      description='Deep learning on Apache Spark with Pytorch',
      keywords = ['pytorch', 'spark', 'sparktorch', 'machine learning', 'deep learning'],
      url='https://github.com/dmmiller612/sparktorch',
      download_url='https://github.com/dmmiller612/sparktorch/archive/0.0.2.tar.gz',
      author='Derek Miller',
      author_email='dmmiller612@gmail.com',
      long_description=open("README.md", "r", encoding='utf-8').read(),
      long_description_content_type="text/markdown",
      install_requires=['torch', 'flask', 'requests', 'dill'],
      license='MIT',
      packages=find_packages(),
      zip_safe=False)



"""
The MIT License

Copyright 2019 Derek Miller

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions
of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

from sparktorch.util import handle_features, load_torch_model, TorchObj, DataObj
from sparktorch.server import Server
import requests
import dill
from typing import Dict, List
from uuid import uuid4


def get_state_dict(master_url: str = 'localhost:3000') -> Dict:
    r = requests.get('http://{0}/parameters'.format(master_url))
    state_dict = dill.loads(r.content)
    return state_dict


def put_deltas_to_server(delta: List, master_url: str = 'localhost:3000'):
    requests.post('http://{0}/update'.format(master_url), data=dill.dumps(delta))


def get_main(master_url: str = 'localhost:3000') -> str:
    r = requests.get('http://{0}/'.format(master_url))
    return r.text


def handle_model(
    data: List[DataObj],
    torch_obj: TorchObj,
    master_url: str = 'localhost:3000',
    iters: int = 1000,
    verbose: int = 1
):

    partition_id = str(uuid4())
    data = list(data)
    if data is None or len(data) == 0:
        return

    data_obj = handle_features(data)

    x_train = data_obj.x_train
    y_train = data_obj.y_train

    torch_obj = load_torch_model(torch_obj)

    model = torch_obj.model
    model.train()

    criterion = torch_obj.criterion

    for i in range(iters):
        state_dict = get_state_dict(master_url)
        model.load_state_dict(state_dict)

        y_pred = model(x_train)
        loss = criterion(y_pred, y_train)
        loss.backward()

        gradients = []
        for param in model.parameters():
            gradients.append(param.grad)

        put_deltas_to_server(gradients)

        if verbose:
            print(f"Partition: {partition_id}. Iteration: {i}. Loss: {loss.item()}")


def train(
    rdd,
    torch_obj: TorchObj,
    server: Server,
    iters: int = 10,
    partition_shuffles: int = 1,
    verbose: int = 1
) -> Dict:
    try:
        master_url = str(server.master_url)

        for i in range(partition_shuffles):
            rdd.foreachPartition(
                lambda x: handle_model(x, torch_obj=torch_obj, master_url=master_url, iters=iters, verbose=verbose)
            )

            if partition_shuffles - i > 1:
                num_partitions = rdd.getNumPartitions()
                rdd = rdd.repartition(num_partitions)

        state_dict = get_state_dict(master_url)
        server.stop_server()

        return state_dict

    except Exception as e:
        server.stop_server()
        raise e




"""Simple reader-writer locks in Python
Many readers can hold the lock XOR one and only one writer
http://majid.info/blog/a-reader-writer-lock-for-python/
"""

import threading

version = """$Id: 04-1.html,v 1.3 2006/12/05 17:45:12 majid Exp $"""


class RWLock:
    """
    A simple reader-writer lock Several readers can hold the lock
    simultaneously, XOR one writer. Write locks have priority over reads to
    prevent write starvation.
    """
    def __init__(self):
        self.rwlock = 0
        self.writers_waiting = 0
        self.monitor = threading.Lock()
        self.readers_ok = threading.Condition(self.monitor)
        self.writers_ok = threading.Condition(self.monitor)

    def acquire_read(self):
        """
        Acquire a read lock. Several threads can hold this typeof lock.
        It is exclusive with write locks.
        """
        self.monitor.acquire()
        while self.rwlock < 0 or self.writers_waiting:
            self.readers_ok.wait()
        self.rwlock += 1
        self.monitor.release()

    def acquire_write(self):
        """
        Acquire a write lock. Only one thread can hold this lock, and
        only when no read locks are also held.
        """
        self.monitor.acquire()
        while self.rwlock != 0:
            self.writers_waiting += 1
            self.writers_ok.wait()
            self.writers_waiting -= 1
        self.rwlock = -1
        self.monitor.release()

    def release(self):
        """
        Release a lock, whether read or write.
        """
        self.monitor.acquire()
        if self.rwlock < 0:
            self.rwlock = 0
        else:
            self.rwlock -= 1
        wake_writers = self.writers_waiting and self.rwlock == 0
        wake_readers = self.writers_waiting == 0
        self.monitor.release()
        if wake_writers:
            self.writers_ok.acquire()
            self.writers_ok.notify()
            self.writers_ok.release()
        elif wake_readers:
            self.readers_ok.acquire()
            self.readers_ok.notifyAll()
            self.readers_ok.release()